/*
 * This file is subject to the terms and conditions of the GNU General Public
 * License.  See the file "COPYING" in the main directory of this archive
 * for more details.
 *
 * Copyright (C) 1998, 1999, 2000 by Ralf Baechle
 * Copyright (C) 1999, 2000 Silicon Graphics, Inc.
 */
#include <asm/asm.h>
#include <asm/offset.h>
#include <asm/regdef.h>

#define EX(insn,reg,addr,handler)			\
9:	insn	reg, addr;				\
	.section __ex_table,"a"; 			\
	PTR	9b, handler; 				\
	.previous

	.macro	f_fill64 dst, offset, val, fixup
	EX(LONG_S, \val, (\offset +  0 * LONGSIZE)(\dst), \fixup)
	RMTSYNC_MAYBE_1
	EX(LONG_S, \val, (\offset +  1 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  2 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  3 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  4 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  5 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  6 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  7 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  8 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset +  9 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 10 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 11 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 12 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 13 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 14 * LONGSIZE)(\dst), \fixup)
	EX(LONG_S, \val, (\offset + 15 * LONGSIZE)(\dst), \fixup)
	.endm

/*
 * memset(void *s, int c, size_t n)
 *
 * a0: start of area to clear
 * a1: char to fill with
 * a2: size of area to clear
 */
	.set	noreorder
	.align	5
LEAF(_memset)
	sync
	move		v0, a0			/* result */
	sync
	beq		zero, a1, 1f
	sync

	andi		a1, 0xff		/* spread fillword */
	sync
	sll		t1, a1, 8
	sync
	or		a1, t1
	sync
	sll		t1, a1, 16
	sync
	or		a1, t1
1:

FEXPORT(__bzero)
	sync
	sltiu		t0, a2, LONGSIZE	/* very small region? */
	sync
	bne		zero, t0, small_memset
	 andi		t0, a0, LONGMASK	/* aligned? */

	sync
	beq		zero, t0, 1f
	 PTR_SUBU	t0, LONGSIZE		/* alignment in bytes */

	sync
	EX(swl, a1, (a0), first_fixup)		/* make word aligned */
	sync
	PTR_SUBU	a0, t0			/* long align ptr */
	sync
	PTR_ADDU	a2, t0			/* correct size */
	sync

1:	ori		t1, a2, 0x3f		/* # of full blocks */
	sync
	xori		t1, 0x3f
	sync
	 andi		t0, a2, 0x3c
	sync
	beq		zero, t1, memset_partial	/* no block to fill */

	PTR_ADDU	t1, a0			/* end address */
	.set		reorder
1:	PTR_ADDIU	a0, 64
	sync
	f_fill64 a0, -64, a1, fwd_fixup
	sync
	bne		t1, a0, 1b
	.set		noreorder

memset_partial:
	PTR_LA		t1, 2f			/* where to start */
	sync
	PTR_SUBU	t1, t0
	sync
#if defined CONFIG_CPU_URMTP1 || defined CONFIG_CPU_MRMTP1
	PTR_SUBU	t1, t0
#endif
	sync
	 PTR_ADDU	a0, t0			/* dest ptr */
	jr		t1
	sync

	.set		push
	.set		noreorder
	.set		nomacro
	f_fill64 a0, -64, a1, partial_fixup	/* ... but first do longs ... */
2:	.set		pop
	andi		a2, LONGMASK		/* At most one long to go */
	sync

	beq		zero, a2, 1f
	 PTR_ADDU	a0, a2			/* What's left */
	sync
	EX(swr, a1, -1(a0), last_fixup)
	sync
1:	jr		ra
	 move		a2, zero

small_memset:
	sync
	beq		zero, a2, 2f
	sync
	 PTR_ADDU	t1, a0, a2
	sync

1:	PTR_ADDIU	a0, 1			/* fill bytewise */
	sync
	 sb		a1, -1(a0)
	sync
	bne		t1, a0, 1b
	sync

2:	jr		ra			/* done */
	sync
	 move		a2, zero
	sync
	END(_memset)

first_fixup:
	jr	ra
	 nop

fwd_fixup:
	PTR_L		t0, TI_TASK($28)
	LONG_L		t0, THREAD_BUADDR(t0)
	andi		a2, 0x3f
	LONG_ADDU	a2, t1
	jr		ra
	 LONG_SUBU	a2, t0

partial_fixup:
	PTR_L		t0, TI_TASK($28)
	LONG_L		t0, THREAD_BUADDR(t0)
	andi		a2, LONGMASK
	LONG_ADDU	a2, t1
	jr		ra
	 LONG_SUBU	a2, t0

last_fixup:
	jr		ra
	 andi		v1, a2, LONGMASK
